#ifndef INTUITION_INTERN_H
#define INTUITION_INTERN_H
/*
	Copyright (C) 1995-2001 AROS - The Amiga Research OS
	$Id: intuition_intern.h,v 1.33 2013/10/22 15:50:10 jacadcaps Exp $

	Desc: Intuitions internal structure
	Lang: english
*/
#ifndef AROS_LIBCALL_H
#   include <aros/libcall.h>
#endif
#ifndef EXEC_EXECBASE_H
#   include <exec/execbase.h>
#endif
#ifndef EXEC_SEMAPHORES_H
#   include <exec/semaphores.h>
#endif
#ifndef EXEC_TASKS_H
#   include <exec/tasks.h>
#endif
#ifndef EXEC_PORTS_H
#   include <exec/ports.h>
#endif
#ifndef EXEC_IO_H
#   include <exec/io.h>
#endif
#ifndef EXEC_TYPES_H
#   include <exec/types.h>
#endif
#ifndef GRAPHICS_GFXBASE_H
#   include <graphics/gfxbase.h>
#endif
#ifndef GRAPHICS_RASTPORT_H
#   include <graphics/rastport.h>
#endif
#ifndef GRAPHICS_REGIONS_H
#   include <graphics/regions.h>
#endif
#ifndef GRAPHICS_CLIP_H
#   include <graphics/clip.h>
#endif
#ifndef GRAPHICS_TEXT_H
#   include <graphics/text.h>
#endif
#ifndef INTUITION_INTUITION_H
#   include <intuition/intuition.h>
#endif
#ifndef INTUITION_INTUITIONBASE_H
#   include <intuition/intuitionbase.h>
#endif
#ifndef INTUITION_CLASSES_H
#   include <intuition/classes.h>
#endif
#ifndef INTUITION_CGHOOKS_H
#   include <intuition/cghooks.h>
#endif
#ifndef INTUITION_SCREENS_H
#   include <intuition/screens.h>
#endif
#ifndef PREFS_ICONTROL_H
#   include <prefs/icontrol.h>
#endif
#ifndef PREFS_INPUT_H
#include <prefs/input.h>
#endif
#ifndef CLIB_ALIB_PROTOS_H
#   include <clib/alib_protos.h>
#endif
#ifdef INTUITION_NOTIFY_SUPPORT
#   include <libraries/screennotify.h>
#   include <libraries/notifyintuition.h>
#endif
#ifndef PROTO_INTUITION_H
#include <proto/intuition.h>
#endif

#include "intuition_debug.h"

#include "inputhandler_actions.h"

#ifdef SKINS
#include "intuition_customize.h"
#include "intuition_internmos.h"
#include "intuition_extend.h"
#include "pointerclassmos.h"
#include "pointers.h"
#endif

#include "requesters.h"

/* Needed for aros_print_not_implemented macro */
#include <aros/debug.h>

#include <aros/asmcall.h>

#ifdef __MORPHOS__
void	dprintf(char *, ...) __attribute__ ((format (printf, 1, 2)));
void * memclr(APTR, ULONG);
#endif

#ifdef __MORPHOS__
void * memclr(APTR, ULONG);
#define bzero(a,b) memclr(a,b)
#else /* __MORPHOS__ */
#define memclr(a,b) bzero(a,b)
#endif /* __MORPHOS__ */

/*
 * min()/max() without macro side effects.
 */
#define max(a,b) \
	({typeof(a) _a = (a); \
	typeof(b) _b = (b);	\
	_a > _b ? _a : _b;})

#define min(a,b) \
	({typeof(a) _a = (a); \
	typeof(b) _b = (b); \
	_a > _b ? _b : _a;})

#define EXTENDWORD(x) x = (LONG)((WORD)x);
#define EXTENDUWORD(x) x = (ULONG)((UWORD)x);

/* SANITY CHECK MACRO */
//#define DEBUG_SANITYCHECK

#ifdef DEBUG_SANITYCHECK
#define SANITY_CHECK(x) if (!((ULONG)x)) {dprintf("Losing sanity in %s line %d\n",__FILE__,__LINE__); return;};
#define SANITY_CHECKR(x,v) if (!((ULONG)x)) {dprintf("Losing sanity in %s line %d\n",__FILE__,__LINE__); return v;};
#else
#define SANITY_CHECK(x) if (!((ULONG)x)) return;
#define SANITY_CHECKR(x,v) if (!((ULONG)x)) return v;
#endif

/* Options */

#define	MENUS_BACKFILL	TRUE
#define FRAME_SIZE       0 /* 0 = 1:1 thin,  1 = 2:1 medres like AmigaOS,  2 = 1:1 thick */

#define GADTOOLSCOMPATIBLE
//enables some gadtools-weirdo-code, MUST be set in both gadtools & intui to work!!!

#ifdef __MORPHOS__
#define ILOCKCHECK(a) ((a->task) && (a->task == IBase->IBaseLockTask))
#else
#define ILOCKCHECK(a) 0
#endif

#ifdef SKINS
#undef GetScreenDrawInfo
#define GetScreenDrawInfo(scr) int_GetScreenDrawInfo(scr)
#undef FreeScreenDrawInfo
#define FreeScreenDrawInfo(scr,dri) ;
/* mui reqs must be always on */
#define MUIREQS
#endif

/* simpleref layers have gadgets redrawn when app calls beginrefresh() */
#define BEGINUPDATEGADGETREFRESH
//#define DAMAGECACHE

#define USEWINDOWLOCK
//#define WINDOWLOCKDEBUG

#ifdef USEWINDOWLOCK
	#ifdef WINDOWLOCKDEBUG
		#define LOCKWINDOW {ObtainSemaphore(&IBase->WindowLock);dprintf("%s/%ld: obtained windowlock\n",__FILE__,__LINE__);};
		#define UNLOCKWINDOW {ReleaseSemaphore(&IBase->WindowLock);dprintf("%s/%ld: released windowlock\n",__FILE__,__LINE__);};
		#define ATTEMPTWINDOWLOCK AttemptSemaphore(&IBase->WindowLock)
	#else
		#define LOCKWINDOW ObtainSemaphore(&IBase->WindowLock);
		#define UNLOCKWINDOW ReleaseSemaphore(&IBase->WindowLock);
		#define ATTEMPTWINDOWLOCK AttemptSemaphore(&IBase->WindowLock)
	#endif
#else
#define LOCKWINDOW
#define UNLOCKWINDOW
#define ATTEMPTWINDOWLOCK
#endif
/* jDc: do NOT disable this! */

//#define USEGADGETLOCK
#ifdef USEGADGETLOCK
#define LOCKGADGET ObtainSemaphore(&IBase->GadgetLock);
#define UNLOCKGADGET ReleaseSemaphore(&IBase->GadgetLock);
#define LOCKWINDOWLAYERS(w) ;
#define UNLOCKWINDOWLAYERS(w) ;
#else
#define LOCKGADGET
#define UNLOCKGADGET
#define LOCKWINDOWLAYERS(w) LockLayerInfo(&w->WScreen->LayerInfo);	\
		if (((struct IntWindow *)(w))->borderlayer) {LockLayer(0,((struct IntWindow *)(w))->borderlayer);};	  \
		if (((struct IntWindow *)(w))->wlayer) {LockLayer(0,((struct IntWindow *)(w))->wlayer);};
#define UNLOCKWINDOWLAYERS(w) if (((struct IntWindow *)(w))->wlayer) {UnlockLayer(((struct IntWindow *)(w))->wlayer);};	\
		if (((struct IntWindow *)(w))->borderlayer) {UnlockLayer(((struct IntWindow *)(w))->borderlayer);}	  \
		UnlockLayerInfo(&w->WScreen->LayerInfo);
#endif

#define WLAYER(w) (((struct IntWindow *)(w))->wlayer)
#define BLAYER(w) (((struct IntWindow *)(w))->borderlayer)

#define LOCKIH ObtainSemaphore(&IBase->InputHandlerLock)
#define UNLOCKIH ReleaseSemaphore(&IBase->InputHandlerLock)

#define GBLANKER (&IBase->BlankerGlobal)

#define TIMEVALWINDOWACTIVATION

/* If PROP_RENDER_OPTIMIZATION is set to 1, propgadget code tries to optimize rendering
   during prop gadget knob movement. Only area of the propgadget that is affected by
   prop gadget knob movement is re-rendered.

   Unfortunately this can fail with some programs, like DOpus 4.x text viewer */

#define PROP_RENDER_OPTIMIZATION 0

#define INTERNAL_BOOPSI  1
#define LIFLG_SUPPORTS_OFFSCREEN_LAYERS 2

#define INTUITIONNAME    "intuition.library"
#define MENUBARLABELCLASS "menubarlabelclass"

#define DEFPUBSCREEN TRUE

#define USE_NEWDISPLAYBEEP 1

#define TITLEBUFFERLEN 255

#define NOCBMPATENTS

#ifdef __MORPHOS__
#if INCLUDE_VERSION < 50

/********************************************************************************/
/* imageclass.h AROS extensions */

#define SYSIA_WithBorder  IA_FGPen	/* default: TRUE */
#define SYSIA_Style       IA_BGPen	/* default: SYSISTYLE_NORMAL */

#define SYSISTYLE_NORMAL   0
#define SYSISTYLE_GADTOOLS 1		/* to get arrow images in gadtools look */

/********************************************************************************/
/* gadgetclass.h AROS extenstions */

/* This method is invoked to learn about the sizing requirements of your class,
   before an object is created. This is AROS specific. */
#define GM_DOMAIN 7
struct gpDomain
{
	STACKULONG		MethodID;   /* GM_DOMAIN */
	struct GadgetInfo * gpd_GInfo;  /* see <intuition/cghooks.h> */
	struct RastPort   * gpd_RPort;  /* RastPort to calculate dimensions for. */
	STACKLONG		gpd_Which;  /* see below */
	struct IBox         gpd_Domain; /* Resulting domain. */
	struct TagItem    * gpd_Attrs;  /* Additional attributes. None defined,
						   yet. */
};

/********************************************************************************/

/* gpd_Which */
#define GDOMAIN_MINIMUM 0 /* Calculate minimum size. */
#define GDOMAIN_NOMINAL 1 /* Calculate nominal size. */
#define GDOMAIN_MAXIMUM 2 /* Calculate maximum size. */

#endif /* INCLUDE_VERSION < 50 */
#endif /* ifdef __MORPHOS__ */

/********************************************************************************/
#define GM_MOVETEST 8
/* this method is used by our draggad to tell if it's OK to move the mouse when window
is near screen boundaries and offscreen is disabled. msg = gpInput*/
//retvals are:
#define MOVETEST_MOVE       0
#define MOVETEST_ADJUSTPOS  1

/* ObtainGIRPort must install a 0 clipregion and
   set scrollx/scrolly of layer to 0. Since this
   will be restored only when ReleaseGIRPort is
   called, we must backup the orig values somewhere */

struct LayerContext
{
	struct Region   *clipregion;
	WORD             scroll_x;
	WORD             scroll_y;
	WORD             nestcount;
};

struct ExtRastPort
{
	struct RastPort      rport;
	struct LayerContext  context;
};

/* Preferences */

struct IScreenModePrefs
{
	ULONG smp_DisplayID;
	UWORD smp_Width;
	UWORD smp_Height;
	UWORD smp_Depth;
	UWORD smp_Control;
};

struct IIControlPrefs
{
	UWORD ic_TimeOut;
	WORD  ic_MetaDrag;
	ULONG ic_Flags;
	UBYTE ic_WBtoFront;
	UBYTE ic_FrontToBack;
	UBYTE ic_ReqTrue;
	UBYTE ic_ReqFalse;
};

struct ISerialPrefs
{
	ULONG sp_Unit0Map;
	ULONG sp_BaudRate;

	ULONG sp_InputBuffer;
	ULONG sp_OutputBuffer;

	ULONG sp_InputHandshake;
	ULONG sp_OutputHandshake;

	ULONG sp_Parity;
	ULONG sp_BitsPerChar;
	ULONG sp_StopBits;
};

struct Color32
{
	ULONG red;
	ULONG green;
	ULONG blue;
};

#define	COLORTABLEENTRIES	32	//8

struct IntScreen;

/* IntuitionBase */
struct IntIntuitionBase
{
	struct IntuitionBase    Base;
#ifdef __MORPHOS__
	WORD                   _MinXMouse,_MaxXMouse;	/* Old 1.3 Base entries*/
	WORD                   _MinYMouse,_MaxYMouse;	/* Old 1.3 Base entries*/
	ULONG                  _StartSecs,_StartMicros;	/* Old 1.3 Base entries*/
	char                   *SystemRequestTitle;	/* written by locale as it seems..what a crappy interface*/
	char                   *WorkbenchTitle;	/* written by locale as it seems..what a crappy interface*/

	/*
	 * savety pad for intuitionbase accesses
	 * probably needs to be smarter
	 */
	UBYTE                   Pad[0x800];
#endif

	/* Put local shit here, invisible for the user */
	struct GfxBase         *GfxBase;
	struct Library         *LayersBase;
	struct ExecBase        *SysBase;
	struct UtilityBase     *UtilBase;
#if !INTERNAL_BOOPSI
	struct Library         *BOOPSIBase;
#endif
	struct Library         *KeymapBase;
	struct Library         *DOSBase;
#ifdef __MORPHOS__
	struct Library         *CyberGfxBase;
	struct Library         *CGXSystemBase;
	//struct Library         *MUIMasterBase;
	struct Library         *CxBase;
#endif
	struct Library         *LocaleBase;
	
	struct Library         *InputBase;
	struct Library         *TimerBase;

	struct MsgPort         *WorkBenchMP;
	struct Screen          *WorkBench;
	struct SignalSemaphore *IBaseLock;

#ifdef __MORPHOS__
	struct Task            *IBaseLockTask;
	LONG                    IBaseLockCount;

	struct Task            *RamlibTask;
#endif

	/* Intuition input handlers replyport. This one is set
	int rom/inputhandler.c/InitIIH()
	*/
	struct MsgPort         *IntuiReplyPort;
	struct MinList         *IntuiActionQueue;
	struct IOStdReq        *InputIO;
	struct MsgPort         InputMP;
	BOOL                    InputDeviceOpen;
	struct Interrupt       *InputHandler;

	struct Hook            *GlobalEditHook;
	/* The default global edit hook */
	struct Hook             DefaultEditHook;


	APTR                    DriverData; /* Pointer which the driver may use */

	struct Screen          *DefaultPubScreen;
	struct SignalSemaphore  PubScrListLock;
	struct MinList          PubScreenList;
	UWORD                   pubScrGlobalMode;

	struct SignalSemaphore  GadgetLock;
	struct SignalSemaphore  MenuLock;
	struct SignalSemaphore  WindowLock;
	struct SignalSemaphore  IntuiActionLock;
	struct SignalSemaphore  InputHandlerLock;
	struct SignalSemaphore  LIBOpenLock;
	struct SignalSemaphore  IDCMPCacheLock;
	struct LayerContext     BackupLayerContext;

	struct IClass          *sizebuttonclass;
	struct IClass          *propgclass;

#ifdef SKINS
	struct IClass          *windowclass;
	struct IClass          *windowsysiclass;
	struct IClass          *screenclass;
	struct IClass          *monitorclass;
	struct IClass          *imsgclass;
#endif

	APTR                   *mosmenuclass;

	struct Preferences      DefaultPreferences;
	struct Preferences      ActivePreferences;

	struct MsgPort         *MenuHandlerPort;
	BOOL                    MenusActive;

	struct TextFont        *ScreenFont;
	struct TextFont        *TopazFont;

	/* Dos function DisplayError() before intuition.library patched it */
	APTR                    OldDisplayErrorFunc;

#if INTERNAL_BOOPSI
	struct SignalSemaphore  ClassListLock;
	struct MinList          ClassList;
	struct IClass           RootClass;
#endif

#ifdef __MORPHOS__
	struct ViewExtra       *ViewLordExtra;
	LONG                    SpriteNum;
#endif

#ifdef SKINS
	ULONG                  *SmallMenuPool;
#endif
	ULONG                  *IDCMPPool;
	ULONG                  *MenutaskPool;

	struct List             IDCMPCacheList;

	struct AmbientScreenDesc AmbientScreenMode;
	struct IIControlPrefs   IControlPrefs;
	struct ISerialPrefs	    SerialPrefs;
#ifdef SKINS
	struct IAction         *IControlActions;
	struct IControlExtensions   IControlExtensions;
	ULONG                   NumIControlActions;
	struct InputPrefsExt    InputPrefsExt;
	struct InputMousePrefs  InputMousePrefs;
	struct InputKeyboardPrefs InputKeyboardPrefs;
	char                    pointermodule[32];
	struct LayerSettings    LayerSettings;
#endif

	struct IClass          *pointerclass;

#ifndef SKINS
	Object                 *DefaultPointer;
	Object                 *BusyPointer;
#else
	Object                 *DefaultPointers[POINTERTYPE_NUMTYPES];
	Object                 *Pointers[POINTERTYPE_NUMTYPES];
	Object                 *RGBPointers[POINTERTYPE_NUMTYPES];
	ULONG                   PointerInitDone;
#endif

	UWORD                   DriPens2[NUMDRIPENS];
	UWORD                   DriPens4[NUMDRIPENS];
	UWORD                   DriPens8[NUMDRIPENS];
	struct Color32          Colors[COLORTABLEENTRIES];
	ULONG                   DMStartSecs;
	ULONG                   DMStartMicro;
	struct IntScreen       *MenuVerifyScreen;

	ULONG                   SizeVerifyCnt;

	ULONG                   PointerDelay;

#ifdef SKINS
	ULONG                   GadgetPointer;
	ULONG                   ClassPointer;
	ULONG                   PointerCheck;
	struct Rectangle        PointerArea;
	struct Window          *PointerWindow;
	ULONG                   PointerGadMode;

	struct SignalSemaphore  DataTypesSem;
	struct Library         *DataTypesBase; /* should be opened ONLY by int_InitCustomChanges!*/
#endif

	ULONG                   LastClickSecs;
	ULONG                   LastClickMicro; /* for doubleclick to front */
	ULONG                   DoubleClickCounter;
	ULONG                   DoubleClickTitleCounter;
	ULONG                   DoubleClickButton;

#ifdef SKINS
	struct Hook             transphook; /* hook for windows with intui transp */
	struct Hook             notransphook; /* hook for windows with no additional transp (borderless,etc) */
#endif

	struct Library         *ScreenNotifyBase;
	struct Library         *NotifyIntuitionBase;

	struct RastPort         DoGadgetMethodRP;
	struct GadgetInfo       DoGadgetMethodGI;

#ifdef USEGETIPREFS
	BOOL                    IPrefsLoaded;
#endif

#if USE_NEWDISPLAYBEEP
	LONG                    BeepingScreens;
#endif

	WORD                    prop_clickoffset_x, prop_clickoffset_y;

#ifdef SKINS
	struct Window          *ActiveToolbarWindow;
	struct Window          *HoveringWindow; /* READ ONLY except for menutask.c!!! use only as TRUE/FALSE indicator in other srcs */

	struct MsgPort         *SkinManagerPort;
	struct MsgPort         *ScreenManagerPort;
	void                   *ScreenManagerHandle;

	struct BlankerGlobal    BlankerGlobal;
	//struct Blanker          Blanker;
	//struct Blanker          BlankerDPMS;
	struct BlankerPrefs	    BlankerPrefs;
	ULONG                   BlankerTimeCnt;
	ULONG                   BlankerMouseHide;
	ULONG                   BlankerMouseTimeCnt;

	struct IMonitorNode    *IMonitorNode;
	struct IMonitorNode    *DefaultMonitorNode;

	struct Complain         LibraryComplain[NUM_COMPLAIN_LIBS];

	struct List             DeadPorts;

	ULONG                   HackTable[10];

	struct EmergencyBoot    EmergencyBoot;

	struct Catalog         *Catalog;

	char                    Codepage[64];
#endif
};

struct SharedPointer
{
	struct ExtSprite *sprite;
	WORD xoffset, yoffset;
	int ref_count;
};

struct SharedPointer *CreateSharedPointer(struct ExtSprite *, int, int, struct IntuitionBase *);
void ObtainSharedPointer(struct SharedPointer *, struct IntuitionBase *);
void ReleaseSharedPointer(struct SharedPointer *, struct IntuitionBase *);
void ObtainSharedPointerNL(struct SharedPointer *, struct IntuitionBase *);
void ReleaseSharedPointerNL(struct SharedPointer *, struct IntuitionBase *);

#ifdef INTUITION_NOTIFY_SUPPORT
void sn_DoNotify(ULONG type,
				 APTR value,
				 struct Library *_ScreenNotifyBase);
#endif

struct IntScreen
{
	struct Screen            Screen;

	/* Private fields */
#ifdef SKINS
	struct IntDrawInfo       DInfo;
#else
	struct DrawInfo          DInfo;
#endif

	ULONG                    realdepth;

	struct TTextAttr         textattr;
	ULONG                    textattrtags[3];
	UWORD                    Pens[NUMDRIPENS];
	struct PubScreenNode    *pubScrNode;
	Object                  *depthgadget;
	UWORD                    SpecialFlags;
	struct Layer            *rootLayer;
	ULONG                    DisplayBeepColor0[3];
	struct Window           *DisplayBeepWindow;
	ULONG                    ModeID;
	struct MonitorSpec      *Monitor;
	struct SharedPointer    *Pointer;
	struct Window           *PointerWindow;
	BOOL 					 ShowPointer;
	BOOL                     SysFont;
	struct Window           *MenuVerifyActiveWindow;
	int                      MenuVerifyTimeOut;
	int                      MenuVerifyMsgCount;
	ULONG                    MenuVerifySeconds;
	ULONG                    MenuVerifyMicros;
	struct BitMap           *AllocatedBitmap;
	struct DBufInfo         *RestoreDBufInfo;

#ifdef SKINS
	WORD                     LastClockPos;
	WORD                     LastClockWidth;
	WORD                     LastTitlePos;

	struct SkinInfo          SkinInfo;
	struct RastPort         *TitlebarBufferRP;
	struct Window           *TitlebarBufferWin;
	ULONG                    TitlebarWinWidth;
	ULONG                    TitlebarWinActive;

	struct IMonitorNode     *IMonitorNode;
	ULONG                    BlankerOffCounter;

	ULONG                    VESAMode;
	ULONG                    VESAModeDepthID;
	ULONG                   *ModeControl;

	struct GammaControl      GammaControl;

	BOOL                     WindowLock;

	struct Window           *STitlebarWindow;
	struct Window           *EventOwnerWindow;
	struct IAction           RecallHotkey;

	BOOL                     Support3D;
	BOOL                     Compositing;
	BOOL                     BlankerScreen;

	void                    *PInfo;

	ULONG                    PixelFormat;

#endif
#if USE_NEWDISPLAYBEEP
	UBYTE                    BeepingCounter;
#endif
};

#define IW(x) ((struct IntWindow *)x)
#define IS(x) ((struct IntScreen *)x)
#define IIM(x) ((struct IntIntuiMessage *)x)

/* SpecialFlags */
#define SF_IsParent	              (0x0001)
#define SF_IsChild	              (0x0002)
#define SF_InvisibleBar           (0x0004)
#define SF_AppearingBar           (0x0008)
#define SF_MOSSpecial             (0x0016)

#ifndef __MORPHOS__

struct IntIntuiMessage
{
	struct ExtIntuiMessage  eimsg;

	/*
	** The following field is needed, because in case of IDCMP_RAWKEY
	** IntuiMessage->IAddress is a pointer to this data, not the data
	** itself (which is the case for IDCMP_VANILLAKEY)
	*/

	APTR                    prevCodeQuals;
};

#endif

#define INT_INTUIMESSAGE(x) ((struct IntIntuiMessage *)(x))

#define TOOLBARWINDOW IBase->ActiveToolbarWindow
#define HOVERINGWINDOW IBase->HoveringWindow

/*extern struct IntuitionBase * IntuitionBase;*/

#define GetPubIBase(ib)   ((struct IntuitionBase *)ib)
#define GetPrivIBase(ib)  ((struct IntIntuitionBase *)ib)
#define IBase GetPrivIBase(IntuitionBase)

#define IDefaultMonitor (IBase->DefaultMonitorNode)
#define IPrimaryMonitor (IBase->IMonitorNode)
#define IsDefaultMonitor(_m) (_m == IDefaultMonitor)

#ifdef GfxBase
#undef GfxBase
#endif
#define _GfxBase     (IBase->GfxBase)
#define GfxBase     _GfxBase

#ifdef LayersBase
#undef LayersBase
#endif
#define _LayersBase     (IBase->LayersBase)
#define LayersBase     _LayersBase

#ifdef SysBase
#undef SysBase
#endif
#define SysBase     (IBase->SysBase)

#ifdef UtilityBase
#undef UtilityBase
#endif
#define UtilityBase (IBase->UtilBase)

#ifdef __MORPHOS__
#ifdef CyberGfxBase
#undef CyberGfxBase
#endif
#define CyberGfxBase (IBase->CyberGfxBase)

#ifdef CGXSystemBase
#undef CGXSystemBase
#endif
#define CGXSystemBase (IBase->CGXSystemBase)

//#ifdef MUIMasterBase
//#undef MUIMasterBase
//#define MUIMasterBase (IBase->MUIMasterBase)
//#endif
#endif

#ifdef LocaleBase
#undef LocaleBase
#endif
#define LocaleBase (IBase->LocaleBase)

#ifdef CxBase
#undef CxBase
#endif
#define CxBase (IBase->CxBase)

#ifdef KeymapBase
#undef KeymapBase
#endif
#define KeymapBase (IBase->KeymapBase)

#ifdef InputBase
#undef InputBase
#endif
#define InputBase (IBase->InputBase)

#ifdef TimerBase
#undef TimerBase
#endif
#define TimerBase (IBase->TimerBase)

#if INTERNAL_BOOPSI

#define BOOPSIBase BOOPSIBase_accessed_although_may_not_use

#else

#ifdef BOOPSIBase
#undef BOOPSIBase
#endif
#define BOOPSIBase (IBase->BOOPSIBase)

#endif


#ifdef DOSBase
#undef DOSBase
#endif
#define DOSBase (IBase->DOSBase)

#define PublicClassList ((struct List *)&(IBase->ClassList))

/* Needed for close() */
#define expunge() \
AROS_LC0(BPTR, expunge, struct IntuitionBase *, IntuitionBase, 3, Intuition)


/* stegerg: one can have sysgadgets outside of window border! All sysgadgets in window
			border must have set GACT_???BORDER and, if they are in a gzz window, also
		GTYP_GZZGADGET */

#define IS_GZZ_GADGET(gad) (((gad)->GadgetType) & GTYP_GZZGADGET)

#define IS_BORDER_GADGET(gad) \
(IS_GZZ_GADGET(gad) \
|| ((gad)->Activation & (GACT_RIGHTBORDER|GACT_LEFTBORDER|GACT_TOPBORDER|GACT_BOTTOMBORDER)))

#define IS_SYS_GADGET(gad) (((gad)->GadgetType) & GTYP_SYSTYPEMASK)

#define IS_BOOPSI_GADGET(gad) (((gad)->GadgetType & GTYP_GTYPEMASK) == GTYP_CUSTOMGADGET)

/*#define IS_BORDER_GADGET(gad) \
	(((gad->GadgetType) & GTYP_SYSGADGET) \
	|| ((gad)->Activation & (GACT_RIGHTBORDER|GACT_LEFTBORDER|GACT_TOPBORDER|GACT_BOTTOMBORDER))) */

#define IS_REQ_GADGET(gad) ((gad)->GadgetType & GTYP_REQGADGET)

#define IS_SCREEN_GADGET(gad) ((gad)->GadgetType & GTYP_SCRGADGET)

#define SET_GI_RPORT(gi, w, req, gad)	\
(gi)->gi_RastPort = (IS_SCREEN_GADGET(gad) ? \
((gi)->gi_Screen->BarLayer ? (gi)->gi_Screen->BarLayer->rp : NULL) : \
(IS_BORDER_GADGET(gad) ?  (w)->BorderRPort : \
((req) ? ((req)->ReqLayer->rp) : (w)->RPort)) \
)


#define REFRESHGAD_BOOPSI	1	/* boopsi gadgets */
#define REFRESHGAD_BORDER	2	/* gadgets in window border */
#define REFRESHGAD_REL		4	/* gadgets with GFLG_RELRIGHT, GFLG_RELBOTTOM,GFLG_RELWIDTH, GFLG_RELHEIGHT */
#define REFRESHGAD_RELS		8	/* GFLG_RELSPECIAL gadgets */
#define REFRESHGAD_TOPBORDER	16	/* used by setwindowtitle */
#define REFRESHGAD_NOGADTOOLS   32  /* used in some cases for _mustbe_ */

#define SYSGADGET_ACTIVE (iihdata->ActiveSysGadget != NULL)

VOID int_refreshglist(struct Gadget *gadgets, struct Window *window,
					  struct Requester *requester, LONG numGad, LONG mustbe, LONG mustnotbe,
					  struct IntuitionBase *IntuitionBase);

VOID int_RefreshWindowFrame(struct Window *window, LONG mustbe, LONG mustnotbe,
							LONG mode,
							struct IntuitionBase *IntuitionBase);

#define int_refreshwindowframe(a,b,c,d) int_RefreshWindowFrame(a,b,c,0,d);

struct ActivateWindowActionMsg
{
	struct IntuiActionMsg msg;
	struct Window *window;
};

VOID int_activatewindow(struct ActivateWindowActionMsg *msg,
						struct IntuitionBase *IntuitionBase);

struct RethinkDisplayActionMsg
{
	struct IntuiActionMsg msg;
	BOOL   lock;
};

void int_RethinkDisplay(struct RethinkDisplayActionMsg *msg,struct IntuitionBase *IntuitionBase);
struct DrawInfo *int_GetScreenDrawInfo(struct Screen *scr);
void SetWindowPointerTypeAsync(ULONG pointertype, struct Window *win, struct IntuitionBase *IntuitionBase);

/* Keep the system gadgets in the same order than the original intuition,
 * some programs make bad asumptions about that...
 */
enum {
	DEPTHGAD,
	ZOOMGAD,
	SIZEGAD,
	CLOSEGAD,
	DRAGBAR,
	ICONIFYGAD,
	LOCKGAD,
	MUIGAD,
	POPUPGAD,
	SNAPSHOTGAD,
	JUMPGAD,
	NUM_SYSGADS
};

#define SYSGAD(w, idx) (((struct IntWindow *)(w))->sysgads[idx])

struct IntWindow
{
	struct Window           window;
#ifdef SKINS
	struct Window          *parent;
#endif

	Object                 *sysgads[NUM_SYSGADS];
	struct Image           *AmigaKey;
	struct Image           *Checkmark;
	struct Window          *menulendwindow;

	/* When the Zoom gadget is pressed the window will have the
	   dimensions stored here. The old dimensions are backed up here
	   again. */
	WORD                    ZipLeftEdge;
	WORD                    ZipTopEdge;
	WORD                    ZipWidth;
	WORD                    ZipHeight;

	WORD                    InitialLeft;
	WORD                    InitialTop;
	WORD                    InitialWidth;
	WORD                    InitialHeight;

	/* max. number of mousemove events to send to this window */
	WORD                    mousequeue;

	/* act. number of mousemove events sent to this window */
	WORD                    num_mouseevents;

	/* max. number of repeated IDCMP_RAWKEY, IDCMP_VANILLAKEY and IDCMP_IDCMPUPDATE
	   messages to send to this window */
	WORD                    repeatqueue;

	/* act. number of repeated IDCMP_RAWKEY, IDCMP_VANILLAKEY and IDCMP_IDCMPUPDATE
	   messages sent to this window */
	WORD                    num_repeatevents;

	/* number of unreplied IDCMP_IDCMPUPDATE messages sent to window
	   used in few hacks :) */
	WORD                    num_idcmpupdate;

#ifdef SKINS
	WORD                    num_sizeevents;
	WORD                    num_refreshevents;
#endif

	WORD                    sizeimage_width;
	WORD                    sizeimage_height;

	ULONG                   helpflags;
	ULONG                   helpgroup;

	Object                 *pointer;
	BOOL                    busy;
	BOOL                    free_pointer;
	UWORD                   pointer_delay;
#ifdef SKINS
	ULONG                   pointertype;
	ULONG                   pointer_delay_pointertype;
	Object                 *pointer_delay_pointer;
	BOOL                    pointer_delay_busy;
#endif

	ULONG                   extrabuttons;
	ULONG                   extrabuttonsid; /* replaces ETI_Dummy if apps wishes it */
#ifdef SKINS
	LONG                    titlepos; /* used by titlebar pattern fill stuff */
	LONG                    leftpos;
	LONG                    rightpos;
#endif
	ULONG                   specialflags;

#ifdef SKINS
	struct Region          *transpregion;
	struct Region          *gzztranspregion;
	struct Hook            *usertransphook;
	struct Region          *usertranspregion;
	ULONG                   useropacity;
	ULONG                   usersourcealpha;
	ULONG                   userhitlevel;
#endif

#ifdef DAMAGECACHE
	struct Region          *trashregion;
#endif

	char                    titlebuffer[TITLEBUFFERLEN+1];

	struct IntuiMessage    *messagecache; //for idcmpupdate cache

	struct Layer           *wlayer;
	struct Layer           *borderlayer;

	struct timeval          lastmsgsent;
	struct timeval          lastmsgreplied;

#ifdef TIMEVALWINDOWACTIVATION
	struct timeval          activationtime;
	ULONG                   clickcnt;
#endif

#ifdef SKINS
	struct Hook             custombackfill;
	struct HookData         hd;
#endif

	struct IntRequestUserData *requserdata;
#ifdef SKINS
	struct Task            *owner;
	ULONG                   ownerid;
	Object                 *Application;
	ULONG                   AspectSizeNumerator;
	ULONG                   AspectSizeDenominator;
	ULONG                   AspectSizeExtraWidth;
	ULONG                   AspectSizeExtraHeight;

	ULONG                   EffectBits;
	BOOL                    DefaultEffectBits; // true if effectbits weren't modified by the user
#endif
};

#define SPFLAG_ICONIFIED	1
#define SPFLAG_NOICONIFY	2
#define SPFLAG_SKININFO		4
#define SPFLAG_LAYERREFRESH	8
#define SPFLAG_TRANSPHOOK	16
#define SPFLAG_LAYERRESIZED 32
#define SPFLAG_USERPORT     64
#define SPFLAG_IAMDEAD      128
#define SPFLAG_CLOSING      256 //used with iamdead
#define SPFLAG_WANTBUFFER   512
#define SPFLAG_TASKISDEAD   1024
#define SPFLAG_DONOTCLOSE   2048

#define HELPF_ISHELPGROUP  1
#define HELPF_GADGETHELP   2

#define IS_NOCAREREFRESH(win) (((win)->Flags & WFLG_NOCAREREFRESH) ? TRUE  : FALSE)
#define IS_DOCAREREFRESH(win) (((win)->Flags & WFLG_NOCAREREFRESH) ? FALSE : TRUE )
#define IS_SIMPLEREFRESH(win) (((win)->Flags & WFLG_SIMPLE_REFRESH) ? TRUE : FALSE)
#define IS_GZZWINDOW(win)     (((win)->Flags & WFLG_GIMMEZEROZERO) ? TRUE  : FALSE)


/* Flag definitions for MoreFlags */

#define WMFLG_NOTIFYDEPTH        (1 << 0)   /* Window wants notification when
it's depth arranged */

#define WMFLG_DO_UNLOCKPUBSCREEN (1 << 1)
#define WMFLG_MENUHELP           (1 << 2)
#define WMFLG_POINTERDELAY       (1 << 3)
#define WMFLG_TABLETMESSAGES     (1 << 4)
#define WMFLG_IAMMUI             (1 << 5)
#define WMFLG_TOOLBARWINDOW      (1 << 6)
#define WMFLG_FRONTWINDOW        (1 << 7)
#define WMFLG_IGNOREME           (1 << 8)
#define WMFLG_THINBORDERS        (1 << 9)
#define WMFLG_NOHIDEONZOOM       (1 << 10) /* do not hide me when zoommaximizing */
#define WMFLG_SELFRESIZING       (1 << 11)

struct IntScreenBuffer
{
	struct ScreenBuffer sb;
	BOOL                free_bitmap;
};

/* Driver prototypes */

extern int  intui_init (struct IntuitionBase *);
extern int  intui_open (struct IntuitionBase *);
extern void intui_close (struct IntuitionBase *);
extern void intui_expunge (struct IntuitionBase *);
extern int intui_GetWindowSize (void);
extern void intui_WindowLimits (struct Window * window,
WORD MinWidth, WORD MinHeight, UWORD MaxWidth, UWORD MaxHeight);
extern void intui_ActivateWindow (struct Window *);
extern BOOL intui_ChangeWindowBox (struct Window * window, WORD x, WORD y,
WORD width, WORD height);
extern void intui_CloseWindow (struct Window *, struct IntuitionBase *);
extern void intui_MoveWindow (struct Window * window, WORD dx, WORD dy);
extern int  intui_OpenWindow (struct Window *, struct IntuitionBase *,
struct BitMap * SuperBitMap, struct Hook *backfillhook,
struct Region * shape,
struct Layer * parent,
ULONG  visible);
extern void intui_SetWindowTitles (struct Window *, UBYTE *, UBYTE *);
extern void intui_RefreshWindowFrame(struct Window *win);
extern struct Window *intui_FindActiveWindow(struct InputEvent *ie, struct IntuitionBase *IntuitionBase);
extern void intui_ScrollWindowRaster(struct Window * win, WORD dx, WORD dy, WORD xmin,
WORD ymin, WORD xmax, WORD ymax, struct IntuitionBase * IntuitionBase);

/* wbtasktalk protos */

ULONG TellWBTaskToCloseWindows(struct IntuitionBase *IntuitionBase);
ULONG TellWBTaskToOpenWindows(struct IntuitionBase *IntuitionBase);

/* intuition_misc protos */
extern void LoadDefaultPreferences(struct IntuitionBase * IntuitionBase);
extern void CheckRectFill(struct RastPort *rp, WORD x1, WORD y1, WORD x2, WORD y2, struct IntuitionBase * IntuitionBase);
extern BOOL CreateWinSysGadgets(struct Window *w, struct IntuitionBase *IntuitionBase);
extern VOID KillWinSysGadgets(struct Window *w, struct IntuitionBase *IntuitionBase);
extern void CreateScreenBar(struct Screen *scr, struct IntuitionBase *IntuitionBase);
extern void KillScreenBar(struct Screen *scr, struct IntuitionBase *IntuitionBase);
extern void RenderScreenBar(struct Screen *scr, BOOL refresh, struct IntuitionBase *IntuitionBase);
extern void UpdateMouseCoords(struct Window *win);
extern WORD SubtractRectFromRect(struct Rectangle *a, struct Rectangle *b, struct Rectangle *destrectarray);

/* misc.c */
extern struct RastPort *MyCreateRastPort(struct IntuitionBase *);
extern struct RastPort *MyCloneRastPort(struct IntuitionBase *, struct RastPort *);
extern void MyFreeRastPort(struct IntuitionBase *, struct RastPort *);
extern struct ExtRastPort *CloneRastPortContext(struct IntuitionBase *,struct RastPort *);
extern void FreeRastPortContext(struct IntuitionBase *, struct ExtRastPort *);
//extern void MySetPointerPos(struct IntuitionBase *, int, int);
struct TextFont *SafeReopenFont(struct IntuitionBase *, struct TextFont **);
extern Object *MakePointerFromPrefs(struct IntuitionBase *, struct Preferences *);
extern Object *MakePointerFromData(struct IntuitionBase *, UWORD *, int, int, int, int,struct PointerColorData *pcd);
void cliprectfill(struct Screen *scr, struct RastPort *rp,WORD x1, WORD y1, WORD x2, WORD y2,struct IntuitionBase *IntuitionBase);
void drawwindowframe(struct Screen *scr, struct RastPort *rp,WORD x1, WORD y1, WORD x2, WORD y2,struct IntuitionBase *IntuitionBase);
void InstallPointer(struct IntuitionBase *, Object **, Object *);
void SetPointerColors(struct IntuitionBase *IntuitionBase);
struct IClass *InitITextIClass (struct IntuitionBase * IntuitionBase);
struct Gadget *DoActivateGadget(struct Window *win, struct Requester *req, struct Gadget *gad,
struct IntuitionBase *IntuitionBase);
VOID DoGMLayout(struct Gadget		*glist,
struct Window		*win,
struct Requester	*req,
UWORD			numgad,
BOOL			initial,
struct IntuitionBase	*IntuitionBase);
BOOL ih_fire_intuimessage(struct Window * w,
ULONG Class,
UWORD Code,
APTR IAddress,
struct IntuitionBase *IntuitionBase);
void NotifyDepthArrangement(struct Window *w, struct IntuitionBase *IntuitionBase);

#ifdef __MORPHOS__
BOOL IsLayerHiddenBySibling(struct Layer *layer, BOOL xx);
#endif

void SetupGInfo(struct GadgetInfo *gi, struct Window *win, struct Requester *req,
struct Gadget *gad, struct IntuitionBase *IntuitionBase);

IPTR Custom_DoMethodA(struct IntuitionBase *, struct Gadget *, Msg);

#ifdef __MORPHOS__
#ifdef DoMethodA
#undef DoMethodA
#endif
#define DoMethodA(x, ...)	(REG_A6=(LONG)IntuitionBase, DoMethodA(x, __VA_ARGS__))
#endif

#define Custom_DoMethodA(x, ...) Custom_DoMethodA(IntuitionBase, x, __VA_ARGS__)

#define HAS_CHILDREN(w) (NULL != w->firstchild)

#ifdef __MORPHOS__
#define DeinitRastPort(rp)	((void)0)
#define CreateRastPort()        MyCreateRastPort(IntuitionBase)
#define CloneRastPort(rp)       MyCloneRastPort(IntuitionBase, rp)
#define FreeRastPort(rp)        MyFreeRastPort(IntuitionBase, rp)
//#define SetPointerPos(x, y)     MySetPointerPos(IntuitionBase, x, y)
#define MouseCoordsRelative()	1
#endif

/* Replacement for dos.library/DisplayError() */
AROS_UFP3(LONG, Intuition_DisplayError,
AROS_UFPA(STRPTR, formatStr , A0),
AROS_UFPA(ULONG , IDCMPFlags, D0),
AROS_UFPA(APTR  , args      , A1));

#ifndef __MORPHOS__
#define POINTERA_SharedPointer  0x80039010
#endif

#ifndef __MORPHOS__
#define dprintf kprintf
#endif

#define DEBUG_ACTIVATEGADGET(x)		;
#define DEBUG_ACTIVATEWINDOW(x)		;
#define DEBUG_ADDCLASS(x)		;
#define DEBUG_ADDGADGET(x)		;
#define DEBUG_ADDGLIST(x)		;
#define DEBUG_ALLOCINTUIMESSAGE(x)	;
#define DEBUG_ALLOCSCREENBUFFER(x)	;
#define DEBUG_ALOHAWORKBENCH(x)		;
#define DEBUG_CHANGEWINDOWBOX(x)	;
#define DEBUG_CLOSESCREEN(x)		;
#define DEBUG_CLOSEWINDOW(x)		;
#define DEBUG_CLOSEWORKBENCH(x)		;
#define DEBUG_DISPOSEOBJECT(x)		;
#define DEBUG_DOGADGETMETHOD(x)		;
#define DEBUG_DRAWBORDER(x)		;
#define DEBUG_FINDCLASS(x)		;
#define DEBUG_FREEICDATA(x)		;
#define DEBUG_FREEINTUIMESSAGE(x)	;
#define DEBUG_FREESCREENBUFFER(x)	;
#define DEBUG_FREESCREENDRAWINFO(x)	;
#define DEBUG_GADGETMOUSE(x)		;
#define DEBUG_GETATTR(x)		;
#define DEBUG_GETDEFAULTPUBSCREEN(x)	;
#define DEBUG_GETDEFPREFS(x)		;
#define DEBUG_GETGADGETIBOX(x)		;
#define DEBUG_GETPREFS(x)		;
#define DEBUG_GETSCREENDATA(x)		;
#define DEBUG_GETSCREENDRAWINFO(x)	;
#define DEBUG_HIDEWINDOW(x)		;
#ifdef __MORPHOS__
#define DEBUG_INIT(x)			if (SysBase->ex_DebugFlags & EXECDEBUGF_INIT) x;
#else
#define DEBUG_INIT(x)			;
#endif
#define	DEBUG_OPEN(x)			;
#define	DEBUG_CLOSE(x)			;
#define DEBUG_INPUTEVENT(x)		;
#define DEBUG_INTREFRESHGLIST(x)	;
#define DEBUG_INTUITEXTLENGTH(x)	;
#define DEBUG_LENDMENUS(x)		;
#define DEBUG_LOCKPUBSCREEN(x)		;
#define DEBUG_MAKECLASS(x)		;
#define DEBUG_MODIFYIDCMP(x)		;
#define DEBUG_NEWOBJECT(x)		;
#define DEBUG_NEXTOBJECT(x)		;
#define DEBUG_OBTAINGIRPORT(x)		;
#define DEBUG_OFFGADGET(x)		;
#define DEBUG_OFFMENU(x)		;
#define DEBUG_ONGADGET(x)		;
#define DEBUG_ONMENU(x)			;
#define DEBUG_OPENSCREEN(x)		;
#define DEBUG_OPENSCREENTAGLIST(x)	;
#define DEBUG_OPENWINDOW(x)		;
#define DEBUG_OPENWINDOWTAGLIST(x)	;
#define DEBUG_OPENWORKBENCH(x)		;
#define DEBUG_POINTER(x)		;
#define DEBUG_PRINTITEXT(x)		;
#define DEBUG_QUERYOVERSCAN(x)		;
#define DEBUG_REFRESH(x)	    ;
#define DEBUG_RELEASEGIRPORT(x)		;
#define DEBUG_REMEMBER(x)		;
#define DEBUG_REMOVEGLIST(x)	    ;
#define DEBUG_REPORTMOUSE(x)		;
#define DEBUG_REQUEST(x)	    ;
#define DEBUG_SCROLLWINDOWRASTER(x)	;
#define DEBUG_SENDINTUIMESSAGE(x)	;
#define DEBUG_SETATTRS(x)		;
#define DEBUG_SETGADGETATTRS(x)		;
#define DEBUG_SETPOINTER(x)		;
#ifdef __MORPHOS__
#define DEBUG_SETIPREFS(x)		if (SysBase->ex_DebugFlags & EXECDEBUGF_INIT) x;
#else
#define DEBUG_SETIPREFS(x)		if (SysBase->ex_DebugFlags & EXECDEBUGF_INIT) x;
#endif
#ifdef __MORPHOS__
#define DEBUG_SETPREFS(x)		if (SysBase->ex_DebugFlags & EXECDEBUGF_INIT) x;
#else
#define DEBUG_SETPREFS(x)		if (SysBase->ex_DebugFlags & EXECDEBUGF_INIT) x;
#endif
#define DEBUG_SHOWWINDOW(x)		;
#define DEBUG_UNLOCKPUBSCREEN(x)	;
#define DEBUG_WINDOWLIMITS(x)		;
#define DEBUG_WINDOWTOBACK(x)		;
#define DEBUG_WINDOWTOFRONT(x)		;
#define DEBUG_ZIPWINDOW(x)		;
#define	DEBUG_VISITOR(x)		;
#define	DEBUG_WORKBENCH(x)		;
#define	DEBUG_LOCKPUBSCREENLIST(x)	;
#define	DEBUG_UNLOCKPUBSCREENLIST(x)	;
#define	DEBUG_RETHINKDISPLAY(x)		;

#endif /* INTUITION_INTERN_H */
